# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Convert molecules into DGLGraphs
#
# pylint: disable= no-member, arguments-differ, invalid-name

from functools import partial
import torch

from dgl import DGLGraph
from rdkit import Chem
from rdkit.Chem import rdmolfiles, rdmolops
from sklearn.neighbors import NearestNeighbors

__all__ = ['mol_to_graph',
           'smiles_to_bigraph',
           'mol_to_bigraph',
           'smiles_to_complete_graph',
           'mol_to_complete_graph',
           'k_nearest_neighbors',
           'mol_to_nearest_neighbor_graph',
           'smiles_to_nearest_neighbor_graph']

# pylint: disable=I1101
def mol_to_graph(mol, graph_constructor, node_featurizer,
                 edge_featurizer, canonical_atom_order, explicit_hydrogens):
    """Convert an RDKit molecule object into a DGLGraph and featurize for it.

    This function can be used to construct any arbitrary ``DGLGraph`` from an
    RDKit molecule instance.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    graph_constructor : callable
        Takes an RDKit molecule as input and returns a DGLGraph
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to
        update ndata for a DGLGraph.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to
        update edata for a DGLGraph.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph.

    Returns
    -------
    g : DGLGraph
        Converted DGLGraph for the molecule

    See Also
    --------
    mol_to_bigraph
    mol_to_complete_graph
    mol_to_nearest_neighbor_graph
    """
    # Whether to have hydrogen atoms as explicit nodes
    if explicit_hydrogens:
        mol = Chem.AddHs(mol)
    else:
        mol = Chem.RemoveHs(mol)

    if canonical_atom_order:
        new_order = rdmolfiles.CanonicalRankAtoms(mol)
        mol = rdmolops.RenumberAtoms(mol, new_order)
    g = graph_constructor(mol)

    if node_featurizer is not None:
        g.ndata.update(node_featurizer(mol))

    if edge_featurizer is not None:
        g.edata.update(edge_featurizer(mol))

    return g

def construct_bigraph_from_mol(mol, add_self_loop=False):
    """Construct a bi-directed DGLGraph with topology only for the molecule.

    The **i** th atom in the molecule, i.e. ``mol.GetAtomWithIdx(i)``, corresponds to the
    **i** th node in the returned DGLGraph.

    The **i** th bond in the molecule, i.e. ``mol.GetBondWithIdx(i)``, corresponds to the
    **(2i)**-th and **(2i+1)**-th edges in the returned DGLGraph. The **(2i)**-th and
    **(2i+1)**-th edges will be separately from **u** to **v** and **v** to **u**, where
    **u** is ``bond.GetBeginAtomIdx()`` and **v** is ``bond.GetEndAtomIdx()``.

    If self loops are added, the last **n** edges will separately be self loops for
    atoms ``0, 1, ..., n-1``.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.

    Returns
    -------
    g : DGLGraph
        Empty bigraph topology of the molecule
    """
    g = DGLGraph()

    # Add nodes
    num_atoms = mol.GetNumAtoms()
    g.add_nodes(num_atoms)

    # Add edges
    src_list = []
    dst_list = []
    num_bonds = mol.GetNumBonds()
    for i in range(num_bonds):
        bond = mol.GetBondWithIdx(i)
        u = bond.GetBeginAtomIdx()
        v = bond.GetEndAtomIdx()
        src_list.extend([u, v])
        dst_list.extend([v, u])
    g.add_edges(src_list, dst_list)

    if add_self_loop:
        nodes = g.nodes()
        g.add_edges(nodes, nodes)

    return g

def mol_to_bigraph(mol, add_self_loop=False,
                   node_featurizer=None,
                   edge_featurizer=None,
                   canonical_atom_order=True,
                   explicit_hydrogens=False):
    """Convert an RDKit molecule object into a bi-directed DGLGraph and featurize for it.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Bi-directed DGLGraph for the molecule

    Examples
    --------
    >>> from rdkit import Chem
    >>> from dgllife.utils import mol_to_bigraph

    >>> mol = Chem.MolFromSmiles('CCO')
    >>> g = mol_to_bigraph(mol)
    >>> print(g)
    DGLGraph(num_nodes=3, num_edges=4,
             ndata_schemes={}
             edata_schemes={})

    We can also initialize node/edge features when constructing graphs.

    >>> import torch
    >>> from rdkit import Chem
    >>> from dgllife.utils import mol_to_bigraph

    >>> def featurize_atoms(mol):
    >>>     feats = []
    >>>     for atom in mol.GetAtoms():
    >>>         feats.append(atom.GetAtomicNum())
    >>>     return {'atomic': torch.tensor(feats).reshape(-1, 1).float()}

    >>> def featurize_bonds(mol):
    >>>     feats = []
    >>>     bond_types = [Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE,
    >>>                   Chem.rdchem.BondType.TRIPLE, Chem.rdchem.BondType.AROMATIC]
    >>>     for bond in mol.GetBonds():
    >>>         btype = bond_types.index(bond.GetBondType())
    >>>         # One bond between atom u and v corresponds to two edges (u, v) and (v, u)
    >>>         feats.extend([btype, btype])
    >>>     return {'type': torch.tensor(feats).reshape(-1, 1).float()}

    >>> mol = Chem.MolFromSmiles('CCO')
    >>> g = mol_to_bigraph(mol, node_featurizer=featurize_atoms,
    >>>                    edge_featurizer=featurize_bonds)
    >>> print(g.ndata['atomic'])
    tensor([[6.],
            [8.],
            [6.]])
    >>> print(g.edata['type'])
    tensor([[0.],
            [0.],
            [0.],
            [0.]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> g = mol_to_bigraph(mol, explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=9, num_edges=16,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    smiles_to_bigraph
    """
    return mol_to_graph(mol, partial(construct_bigraph_from_mol, add_self_loop=add_self_loop),
                        node_featurizer, edge_featurizer,
                        canonical_atom_order, explicit_hydrogens)

def smiles_to_bigraph(smiles, add_self_loop=False,
                      node_featurizer=None,
                      edge_featurizer=None,
                      canonical_atom_order=True,
                      explicit_hydrogens=False):
    """Convert a SMILES into a bi-directed DGLGraph and featurize for it.

    Parameters
    ----------
    smiles : str
        String of SMILES
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Bi-directed DGLGraph for the molecule

    Examples
    --------
    >>> from dgllife.utils import smiles_to_bigraph

    >>> g = smiles_to_bigraph('CCO')
    >>> print(g)
    DGLGraph(num_nodes=3, num_edges=4,
             ndata_schemes={}
             edata_schemes={})

    We can also initialize node/edge features when constructing graphs.

    >>> import torch
    >>> from rdkit import Chem
    >>> from dgllife.utils import smiles_to_bigraph

    >>> def featurize_atoms(mol):
    >>>     feats = []
    >>>     for atom in mol.GetAtoms():
    >>>         feats.append(atom.GetAtomicNum())
    >>>     return {'atomic': torch.tensor(feats).reshape(-1, 1).float()}

    >>> def featurize_bonds(mol):
    >>>     feats = []
    >>>     bond_types = [Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE,
    >>>                   Chem.rdchem.BondType.TRIPLE, Chem.rdchem.BondType.AROMATIC]
    >>>     for bond in mol.GetBonds():
    >>>         btype = bond_types.index(bond.GetBondType())
    >>>         # One bond between atom u and v corresponds to two edges (u, v) and (v, u)
    >>>         feats.extend([btype, btype])
    >>>     return {'type': torch.tensor(feats).reshape(-1, 1).float()}

    >>> g = smiles_to_bigraph('CCO', node_featurizer=featurize_atoms,
    >>>                       edge_featurizer=featurize_bonds)
    >>> print(g.ndata['atomic'])
    tensor([[6.],
            [8.],
            [6.]])
    >>> print(g.edata['type'])
    tensor([[0.],
            [0.],
            [0.],
            [0.]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> g = smiles_to_bigraph('CCO', explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=9, num_edges=16,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    mol_to_bigraph
    """
    mol = Chem.MolFromSmiles(smiles)
    return mol_to_bigraph(mol, add_self_loop, node_featurizer,
                          edge_featurizer, canonical_atom_order, explicit_hydrogens)

def construct_complete_graph_from_mol(mol, add_self_loop=False):
    """Construct a complete graph with topology only for the molecule

    The **i** th atom in the molecule, i.e. ``mol.GetAtomWithIdx(i)``, corresponds to the
    **i** th node in the returned DGLGraph.

    The edges are in the order of (0, 0), (1, 0), (2, 0), ... (0, 1), (1, 1), (2, 1), ...
    If self loops are not created, we will not have (0, 0), (1, 1), ...

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.

    Returns
    -------
    g : DGLGraph
        Empty complete graph topology of the molecule
    """
    num_atoms = mol.GetNumAtoms()
    edge_list = []
    for i in range(num_atoms):
        for j in range(num_atoms):
            if i != j or add_self_loop:
                edge_list.append((i, j))
    g = DGLGraph(edge_list)

    return g

def mol_to_complete_graph(mol, add_self_loop=False,
                          node_featurizer=None,
                          edge_featurizer=None,
                          canonical_atom_order=True,
                          explicit_hydrogens=False):
    """Convert an RDKit molecule into a complete DGLGraph and featurize for it.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Complete DGLGraph for the molecule

    Examples
    --------
    >>> from rdkit import Chem
    >>> from dgllife.utils import mol_to_complete_graph

    >>> mol = Chem.MolFromSmiles('CCO')
    >>> g = mol_to_complete_graph(mol)
    >>> print(g)
    DGLGraph(num_nodes=3, num_edges=6,
             ndata_schemes={}
             edata_schemes={})

    We can also initialize node/edge features when constructing graphs.

    >>> import torch
    >>> from rdkit import Chem
    >>> from dgllife.utils import mol_to_complete_graph
    >>> from functools import partial

    >>> def featurize_atoms(mol):
    >>>     feats = []
    >>>     for atom in mol.GetAtoms():
    >>>         feats.append(atom.GetAtomicNum())
    >>>     return {'atomic': torch.tensor(feats).reshape(-1, 1).float()}

    >>> def featurize_edges(mol, add_self_loop=False):
    >>>     feats = []
    >>>     num_atoms = mol.GetNumAtoms()
    >>>     atoms = list(mol.GetAtoms())
    >>>     distance_matrix = Chem.GetDistanceMatrix(mol)
    >>>     for i in range(num_atoms):
    >>>         for j in range(num_atoms):
    >>>             if i != j or add_self_loop:
    >>>                 feats.append(float(distance_matrix[i, j]))
    >>>     return {'dist': torch.tensor(feats).reshape(-1, 1).float()}

    >>> mol = Chem.MolFromSmiles('CCO')
    >>> add_self_loop = True
    >>> g = mol_to_complete_graph(
    >>>         mol, add_self_loop=add_self_loop, node_featurizer=featurize_atoms,
    >>>         edge_featurizer=partial(featurize_edges, add_self_loop=add_self_loop))
    >>> print(g.ndata['atomic'])
    tensor([[6.],
            [8.],
            [6.]])
    >>> print(g.edata['dist'])
    tensor([[0.],
            [2.],
            [1.],
            [2.],
            [0.],
            [1.],
            [1.],
            [1.],
            [0.]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> g = mol_to_complete_graph(mol, explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=9, num_edges=72,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    smiles_to_complete_graph
    """
    return mol_to_graph(mol,
                        partial(construct_complete_graph_from_mol, add_self_loop=add_self_loop),
                        node_featurizer, edge_featurizer,
                        canonical_atom_order, explicit_hydrogens)

def smiles_to_complete_graph(smiles, add_self_loop=False,
                             node_featurizer=None,
                             edge_featurizer=None,
                             canonical_atom_order=True,
                             explicit_hydrogens=False):
    """Convert a SMILES into a complete DGLGraph and featurize for it.

    Parameters
    ----------
    smiles : str
        String of SMILES
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Complete DGLGraph for the molecule

    Examples
    --------
    >>> from dgllife.utils import smiles_to_complete_graph

    >>> g = smiles_to_complete_graph('CCO')
    >>> print(g)
    DGLGraph(num_nodes=3, num_edges=6,
             ndata_schemes={}
             edata_schemes={})

    We can also initialize node/edge features when constructing graphs.

    >>> import torch
    >>> from rdkit import Chem
    >>> from dgllife.utils import smiles_to_complete_graph
    >>> from functools import partial

    >>> def featurize_atoms(mol):
    >>>     feats = []
    >>>     for atom in mol.GetAtoms():
    >>>         feats.append(atom.GetAtomicNum())
    >>>     return {'atomic': torch.tensor(feats).reshape(-1, 1).float()}

    >>> def featurize_edges(mol, add_self_loop=False):
    >>>     feats = []
    >>>     num_atoms = mol.GetNumAtoms()
    >>>     atoms = list(mol.GetAtoms())
    >>>     distance_matrix = Chem.GetDistanceMatrix(mol)
    >>>     for i in range(num_atoms):
    >>>         for j in range(num_atoms):
    >>>             if i != j or add_self_loop:
    >>>                 feats.append(float(distance_matrix[i, j]))
    >>>     return {'dist': torch.tensor(feats).reshape(-1, 1).float()}

    >>> add_self_loop = True
    >>> g = smiles_to_complete_graph(
    >>>         'CCO', add_self_loop=add_self_loop, node_featurizer=featurize_atoms,
    >>>         edge_featurizer=partial(featurize_edges, add_self_loop=add_self_loop))
    >>> print(g.ndata['atomic'])
    tensor([[6.],
            [8.],
            [6.]])
    >>> print(g.edata['dist'])
    tensor([[0.],
            [2.],
            [1.],
            [2.],
            [0.],
            [1.],
            [1.],
            [1.],
            [0.]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> g = smiles_to_complete_graph('CCO', explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=9, num_edges=72,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    mol_to_complete_graph
    """
    mol = Chem.MolFromSmiles(smiles)
    return mol_to_complete_graph(mol, add_self_loop, node_featurizer,
                                 edge_featurizer, canonical_atom_order, explicit_hydrogens)

def k_nearest_neighbors(coordinates, neighbor_cutoff, max_num_neighbors=None,
                        p_distance=2, self_loops=False):
    """Find k nearest neighbors for each atom

    We do not guarantee that the edges are sorted according to the distance
    between atoms.

    Parameters
    ----------
    coordinates : numpy.ndarray of shape (N, D)
        The coordinates of atoms in the molecule. N for the number of atoms
        and D for the dimensions of the coordinates.
    neighbor_cutoff : float
        If the distance between a pair of nodes is larger than neighbor_cutoff,
        they will not be considered as neighboring nodes.
    max_num_neighbors : int or None.
        If not None, then this specifies the maximum number of neighbors
        allowed for each atom. Default to None.
    p_distance : int
        We compute the distance between neighbors using Minkowski (:math:`l_p`)
        distance. When ``p_distance = 1``, Minkowski distance is equivalent to
        Manhattan distance. When ``p_distance = 2``, Minkowski distance is
        equivalent to the standard Euclidean distance. Default to 2.
    self_loops : bool
        Whether to allow a node to be its own neighbor. Default to False.

    Returns
    -------
    srcs : list of int
        Source nodes.
    dsts : list of int
        Destination nodes, corresponding to ``srcs``.
    distances : list of float
        Distances between the end nodes, corresponding to ``srcs`` and ``dsts``.

    Examples
    --------
    >>> from dgllife.utils import get_mol_3d_coordinates, k_nearest_neighbors
    >>> from rdkit import Chem
    >>> from rdkit.Chem import AllChem

    >>> mol = Chem.MolFromSmiles('CC1(C(N2C(S1)C(C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C')
    >>> AllChem.EmbedMolecule(mol)
    >>> AllChem.MMFFOptimizeMolecule(mol)
    >>> coords = get_mol_3d_coordinates(mol)
    >>> srcs, dsts, dists = k_nearest_neighbors(coords, neighbor_cutoff=1.25)
    >>> print(srcs)
    [8, 7, 11, 10, 20, 19]
    >>> print(dsts)
    [7, 8, 10, 11, 19, 20]
    >>> print(dists)
    [1.2084666104583117, 1.2084666104583117, 1.226457824344217,
     1.226457824344217, 1.2230522248065987, 1.2230522248065987]

    See Also
    --------
    get_mol_3d_coordinates
    mol_to_nearest_neighbor_graph
    smiles_to_nearest_neighbor_graph
    """
    num_atoms = coordinates.shape[0]
    model = NearestNeighbors(radius=neighbor_cutoff, p=p_distance)
    model.fit(coordinates)
    dists_, nbrs = model.radius_neighbors(coordinates)
    srcs, dsts, dists = [], [], []
    for i in range(num_atoms):
        dists_i = dists_[i].tolist()
        nbrs_i = nbrs[i].tolist()
        if not self_loops:
            dists_i.remove(0)
            nbrs_i.remove(i)
        if max_num_neighbors is not None and len(nbrs_i) > max_num_neighbors:
            packed_nbrs = list(zip(dists_i, nbrs_i))
            # Sort neighbors based on distance from smallest to largest
            packed_nbrs.sort(key=lambda tup: tup[0])
            dists_i, nbrs_i = map(list, zip(*packed_nbrs))
            dsts.extend([i for _ in range(max_num_neighbors)])
            srcs.extend(nbrs_i[:max_num_neighbors])
            dists.extend(dists_i[:max_num_neighbors])
        else:
            dsts.extend([i for _ in range(len(nbrs_i))])
            srcs.extend(nbrs_i)
            dists.extend(dists_i)

    return srcs, dsts, dists

# pylint: disable=E1102
def mol_to_nearest_neighbor_graph(mol,
                                  coordinates,
                                  neighbor_cutoff,
                                  max_num_neighbors=None,
                                  p_distance=2,
                                  add_self_loop=False,
                                  node_featurizer=None,
                                  edge_featurizer=None,
                                  canonical_atom_order=True,
                                  keep_dists=False,
                                  dist_field='dist',
                                  explicit_hydrogens=False):
    """Convert an RDKit molecule into a nearest neighbor graph and featurize for it.

    Different from bigraph and complete graph, the nearest neighbor graph
    may not be symmetric since i is the closest neighbor of j does not
    necessarily suggest the other way.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule holder
    coordinates : numpy.ndarray of shape (N, D)
        The coordinates of atoms in the molecule. N for the number of atoms
        and D for the dimensions of the coordinates.
    neighbor_cutoff : float
        If the distance between a pair of nodes is larger than neighbor_cutoff,
        they will not be considered as neighboring nodes.
    max_num_neighbors : int or None.
        If not None, then this specifies the maximum number of neighbors
        allowed for each atom. Default to None.
    p_distance : int
        We compute the distance between neighbors using Minkowski (:math:`l_p`)
        distance. When ``p_distance = 1``, Minkowski distance is equivalent to
        Manhattan distance. When ``p_distance = 2``, Minkowski distance is
        equivalent to the standard Euclidean distance. Default to 2.
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    keep_dists : bool
        Whether to store the distance between neighboring atoms in ``edata`` of the
        constructed DGLGraphs. Default to False.
    dist_field : str
        Field for storing distance between neighboring atoms in ``edata``. This comes
        into effect only when ``keep_dists=True``. Default to ``'dist'``.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Nearest neighbor DGLGraph for the molecule

    Examples
    --------
    >>> from dgllife.utils import mol_to_nearest_neighbor_graph
    >>> from rdkit import Chem
    >>> from rdkit.Chem import AllChem

    >>> mol = Chem.MolFromSmiles('CC1(C(N2C(S1)C(C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C')
    >>> AllChem.EmbedMolecule(mol)
    >>> AllChem.MMFFOptimizeMolecule(mol)
    >>> coords = get_mol_3d_coordinates(mol)
    >>> g = mol_to_nearest_neighbor_graph(mol, coords, neighbor_cutoff=1.25)
    >>> print(g)
    DGLGraph(num_nodes=23, num_edges=6,
             ndata_schemes={}
             edata_schemes={})

    Quite often we will want to use the distance between end atoms of edges, this can be
    achieved with

    >>> g = mol_to_nearest_neighbor_graph(mol, coords, neighbor_cutoff=1.25, keep_dists=True)
    >>> print(g.edata['dist'])
    tensor([[1.2024],
            [1.2024],
            [1.2270],
            [1.2270],
            [1.2259],
            [1.2259]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> mol = Chem.MolFromSmiles('CC1(C(N2C(S1)C(C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C')
    >>> mol = Chem.AddHs(mol)
    >>> AllChem.EmbedMolecule(mol)
    >>> AllChem.MMFFOptimizeMolecule(mol)
    >>> coords = get_mol_3d_coordinates(mol)
    >>> g = mol_to_nearest_neighbor_graph(mol, coords, neighbor_cutoff=1.25,
    >>>                                   explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=41, num_edges=42,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    get_mol_3d_coordinates
    k_nearest_neighbors
    smiles_to_nearest_neighbor_graph
    """
    if explicit_hydrogens:
        mol = Chem.AddHs(mol)
    else:
        mol = Chem.RemoveHs(mol)

    num_atoms = mol.GetNumAtoms()
    num_coords = coordinates.shape[0]
    assert num_atoms == num_coords, \
        'Expect the number of atoms to match the first dimension of coordinates, ' \
        'got {:d} and {:d}'.format(num_atoms, num_coords)

    if canonical_atom_order:
        new_order = rdmolfiles.CanonicalRankAtoms(mol)
        mol = rdmolops.RenumberAtoms(mol, new_order)

    srcs, dsts, dists = k_nearest_neighbors(coordinates=coordinates,
                                            neighbor_cutoff=neighbor_cutoff,
                                            max_num_neighbors=max_num_neighbors,
                                            p_distance=p_distance,
                                            self_loops=add_self_loop)
    g = DGLGraph()

    # Add nodes first since some nodes may be completely isolated
    g.add_nodes(num_atoms)

    # Add edges
    g.add_edges(srcs, dsts)

    if node_featurizer is not None:
        g.ndata.update(node_featurizer(mol))

    if edge_featurizer is not None:
        g.edata.update(edge_featurizer(mol))

    if keep_dists:
        assert dist_field not in g.edata, \
            'Expect {} to be reserved for distance between neighboring atoms.'
        g.edata[dist_field] = torch.tensor(dists).float().reshape(-1, 1)

    return g

def smiles_to_nearest_neighbor_graph(smiles,
                                     coordinates,
                                     neighbor_cutoff,
                                     max_num_neighbors=None,
                                     p_distance=2,
                                     add_self_loop=False,
                                     node_featurizer=None,
                                     edge_featurizer=None,
                                     canonical_atom_order=True,
                                     keep_dists=False,
                                     dist_field='dist',
                                     explicit_hydrogens=False):
    """Convert a SMILES into a nearest neighbor graph and featurize for it.

    Different from bigraph and complete graph, the nearest neighbor graph
    may not be symmetric since i is the closest neighbor of j does not
    necessarily suggest the other way.

    Parameters
    ----------
    smiles : str
        String of SMILES
    coordinates : numpy.ndarray of shape (N, D)
        The coordinates of atoms in the molecule. N for the number of atoms
        and D for the dimensions of the coordinates.
    neighbor_cutoff : float
        If the distance between a pair of nodes is larger than neighbor_cutoff,
        they will not be considered as neighboring nodes.
    max_num_neighbors : int or None.
        If not None, then this specifies the maximum number of neighbors
        allowed for each atom. Default to None.
    p_distance : int
        We compute the distance between neighbors using Minkowski (:math:`l_p`)
        distance. When ``p_distance = 1``, Minkowski distance is equivalent to
        Manhattan distance. When ``p_distance = 2``, Minkowski distance is
        equivalent to the standard Euclidean distance. Default to 2.
    add_self_loop : bool
        Whether to add self loops in DGLGraphs. Default to False.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    canonical_atom_order : bool
        Whether to use a canonical order of atoms returned by RDKit. Setting it
        to true might change the order of atoms in the graph constructed. Default
        to True.
    keep_dists : bool
        Whether to store the distance between neighboring atoms in ``edata`` of the
        constructed DGLGraphs. Default to False.
    dist_field : str
        Field for storing distance between neighboring atoms in ``edata``. This comes
        into effect only when ``keep_dists=True``. Default to ``'dist'``.
    explicit_hydrogens : bool
        Whether to explicitly represent hydrogens as nodes in the graph. Default to False.

    Returns
    -------
    g : DGLGraph
        Nearest neighbor DGLGraph for the molecule

    Examples
    --------
    >>> from dgllife.utils import smiles_to_nearest_neighbor_graph
    >>> from rdkit import Chem
    >>> from rdkit.Chem import AllChem

    >>> smiles = 'CC1(C(N2C(S1)C(C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C'
    >>> mol = Chem.MolFromSmiles(smiles)
    >>> AllChem.EmbedMolecule(mol)
    >>> AllChem.MMFFOptimizeMolecule(mol)
    >>> coords = get_mol_3d_coordinates(mol)
    >>> g = mol_to_nearest_neighbor_graph(mol, coords, neighbor_cutoff=1.25)
    >>> print(g)
    DGLGraph(num_nodes=23, num_edges=6,
             ndata_schemes={}
             edata_schemes={})

    Quite often we will want to use the distance between end atoms of edges, this can be
    achieved with

    >>> g = smiles_to_nearest_neighbor_graph(smiles, coords, neighbor_cutoff=1.25, keep_dists=True)
    >>> print(g.edata['dist'])
    tensor([[1.2024],
            [1.2024],
            [1.2270],
            [1.2270],
            [1.2259],
            [1.2259]])

    By default, we do not explicitly represent hydrogens as nodes, which can be done as follows.

    >>> smiles = 'CC1(C(N2C(S1)C(C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C'
    >>> mol = Chem.MolFromSmiles(smiles)
    >>> mol = Chem.AddHs(mol)
    >>> AllChem.EmbedMolecule(mol)
    >>> AllChem.MMFFOptimizeMolecule(mol)
    >>> coords = get_mol_3d_coordinates(mol)
    >>> g = smiles_to_nearest_neighbor_graph(smiles, coords, neighbor_cutoff=1.25,
    >>>                                      explicit_hydrogens=True)
    >>> print(g)
    DGLGraph(num_nodes=41, num_edges=42,
             ndata_schemes={}
             edata_schemes={})

    See Also
    --------
    get_mol_3d_coordinates
    k_nearest_neighbors
    mol_to_nearest_neighbor_graph
    """
    mol = Chem.MolFromSmiles(smiles)
    return mol_to_nearest_neighbor_graph(
        mol, coordinates, neighbor_cutoff, max_num_neighbors, p_distance,
        add_self_loop, node_featurizer, edge_featurizer, canonical_atom_order,
        keep_dists, dist_field, explicit_hydrogens)



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Dataset for inference on smiles

from rdkit import Chem

from ..utils.mol_to_graph import mol_to_bigraph

__all__ = ['UnlabeledSMILES']

class UnlabeledSMILES(object):
    """Construct a SMILES dataset without labels for inference.

    We will 1) Filter out invalid SMILES strings and record canonical SMILES strings
    for valid ones 2) Construct a DGLGraph for each valid one and feature its node/edge

    Parameters
    ----------
    smiles_list : list of str
        List of SMILES strings
    mol_to_graph: callable, rdkit.Chem.rdchem.Mol -> DGLGraph
        A function turning an RDKit molecule object into a DGLGraph.
        Default to :func:`dgllife.utils.mol_to_bigraph`.
    node_featurizer : None or callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : None or callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    log_every : bool
        Print a message every time ``log_every`` molecules are processed. Default to 1000.
    """
    def __init__(self, smiles_list, mol_to_graph=mol_to_bigraph, node_featurizer=None,
                 edge_featurizer=None, log_every=1000):
        super(UnlabeledSMILES, self).__init__()

        canonical_smiles = []
        mol_list = []
        for smi in smiles_list:
            mol = Chem.MolFromSmiles(smi)
            if mol is None:
                continue
            mol_list.append(mol)
            canonical_smiles.append(Chem.MolToSmiles(mol))

        self.smiles = canonical_smiles
        self.graphs = []
        for i, mol in enumerate(mol_list):
            if (i + 1) % log_every == 0:
                print('Processing molecule {:d}/{:d}'.format(i + 1, len(self)))
            self.graphs.append(mol_to_graph(mol, node_featurizer=node_featurizer,
                                            edge_featurizer=edge_featurizer))

    def __getitem__(self, item):
        """Get datapoint with index

        Parameters
        ----------
        item : int
            Datapoint index

        Returns
        -------
        str
            SMILES for the ith datapoint
        DGLGraph
            DGLGraph for the ith datapoint
        """
        return self.smiles[item], self.graphs[item]

    def __len__(self):
        """Size for the dataset

        Returns
        -------
        int
            Size for the dataset
        """
        return len(self.smiles)



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# ESOL from MoleculeNet for the prediction of water solubility

import pandas as pd

from dgl.data.utils import get_download_dir, download, _get_dgl_url, extract_archive

from .csv_dataset import MoleculeCSVDataset
from ..utils.mol_to_graph import smiles_to_bigraph

__all__ = ['ESOL']

class ESOL(MoleculeCSVDataset):
    r"""ESOL from MoleculeNet for the prediction of water solubility

    ESOL (delaney) is a standard regression dataset containing structures and water solubility
    data for 1128 compounds. It uses measured log solubility in mols per litre of compounds
    as labels. The dataset is widely used to validate machine learning models on estimating
    solubility directly from molecular structures. Note that these structures don't include
    3D coordinates, since solubility is a property of a molecule and not of its particular
    conformers.

    References:

        * [1] MoleculeNet: A Benchmark for Molecular Machine Learning.
        * [2] ESOL: estimating aqueous solubility directly from molecular structure.

    Parameters
    ----------
    smiles_to_graph: callable, str -> DGLGraph
        A function turning a SMILES string into a DGLGraph.
        Default to :func:`dgllife.utils.smiles_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    log_every : bool
        Print a message every time ``log_every`` molecules are processed. Default to 1000.
    cache_file_path : str
        Path to the cached DGLGraphs, default to 'esol_dglgraph.bin'.

    Examples
    --------

    >>> from dgllife.data import ESOL
    >>> from dgllife.utils import smiles_to_bigraph, CanonicalAtomFeaturizer

    >>> dataset = ESOL(smiles_to_bigraph, CanonicalAtomFeaturizer())
    >>> # Get size of the dataset
    >>> len(dataset)
    1128
    >>> # Get the 0th datapoint, consisting of SMILES, DGLGraph and solubility
    >>> dataset[0]
    ('OCC3OC(OCC2OC(OC(C#N)c1ccccc1)C(O)C(O)C2O)C(O)C(O)C3O ',
     DGLGraph(num_nodes=32, num_edges=68,
              ndata_schemes={}
              edata_schemes={}),
     tensor([-0.7700]))

    We also provide information for the name, estimated solubility, minimum atom
    degree, molecular weight, number of h bond donors, number of rings,
    number of rotatable bonds, and polar surface area of the compound

    >>> # Access the information mentioned above for the ith datapoint
    >>> dataset.compound_names[i]
    >>> dataset.estimated_solubility[i]
    >>> dataset.min_degree[i]
    >>> dataset.mol_weight[i]
    >>> dataset.num_h_bond_donors[i]
    >>> dataset.num_rings[i]
    >>> dataset.num_rotatable_bonds[i]
    >>> dataset.polar_surface_area[i]

    We can also get all these information along with SMILES, DGLGraph and solubility
    at once.

    >>> dataset.load_full = True
    >>> dataset[0]
    ('OCC3OC(OCC2OC(OC(C#N)c1ccccc1)C(O)C(O)C2O)C(O)C(O)C3O ',
     DGLGraph(num_nodes=32, num_edges=68,
              ndata_schemes={}
              edata_schemes={}),
     tensor([-0.7700]),
     'Amigdalin',
     -0.974,
     1,
     457.43200000000013,
     7,
     3,
     7,
     202.32)
    """
    def __init__(self,
                 smiles_to_graph=smiles_to_bigraph,
                 node_featurizer=None,
                 edge_featurizer=None,
                 load=True,
                 log_every=1000,
                 cache_file_path='esol_dglgraph.bin'):

        self._url = 'dataset/ESOL.zip'
        data_path = get_download_dir() + '/ESOL.zip'
        dir_path = get_download_dir() + '/ESOL'
        download(_get_dgl_url(self._url), path=data_path)
        extract_archive(data_path, dir_path)
        df = pd.read_csv(dir_path + '/delaney-processed.csv')

        # Compound names in PubChem
        self.compound_names = df['Compound ID'].tolist()
        # Estimated solubility
        self.estimated_solubility = df['ESOL predicted log solubility in mols per litre'].tolist()
        # Minimum atom degree
        self.min_degree = df['Minimum Degree'].tolist()
        # Molecular weight
        self.mol_weight = df['Molecular Weight'].tolist()
        # Number of H-Bond Donors
        self.num_h_bond_donors = df['Number of H-Bond Donors'].tolist()
        # Number of rings
        self.num_rings = df['Number of Rings'].tolist()
        # Number of rotatable bonds
        self.num_rotatable_bonds = df['Number of Rotatable Bonds'].tolist()
        # Polar Surface Area
        self.polar_surface_area = df['Polar Surface Area'].tolist()

        self.load_full = False

        super(ESOL, self).__init__(df=df,
                                   smiles_to_graph=smiles_to_graph,
                                   node_featurizer=node_featurizer,
                                   edge_featurizer=edge_featurizer,
                                   smiles_column='smiles',
                                   cache_file_path=cache_file_path,
                                   task_names=['measured log solubility in mols per litre'],
                                   load=load,
                                   log_every=log_every,
                                   init_mask=False)

    def __getitem__(self, item):
        """Get datapoint with index

        Parameters
        ----------
        item : int
            Datapoint index

        Returns
        -------
        str
            SMILES for the ith datapoint
        DGLGraph
            DGLGraph for the ith datapoint
        Tensor of dtype float32 and shape (1)
            Labels of the ith datapoint
        str, optional
            Name for the ith compound, returned only when ``self.load_full`` is True.
        float, optional
            Estimated solubility for the ith compound,
            returned only when ``self.load_full`` is True.
        int, optional
            Minimum atom degree for the ith datapoint, returned only when
            ``self.load`` is True.
        float, optional
            Molecular weight for the ith datapoint, returned only when
            ``self.load`` is True.
        int, optional
            Number of h bond donors for the ith datapoint, returned only when
            ``self.load`` is True.
        int, optional
            Number of rings in the ith datapoint, returned only when
            ``self.load`` is True.
        int, optional
            Number of rotatable bonds in the ith datapoint, returned only when
            ``self.load`` is True.
        float, optional
            Polar surface area for the ith datapoint, returned only when
            ``self.load`` is True.
        """
        if self.load_full:
            return self.smiles[item], self.graphs[item], self.labels[item], \
                   self.compound_names[item], self.estimated_solubility[item], \
                   self.min_degree[item], self.mol_weight[item], \
                   self.num_h_bond_donors[item], self.num_rings[item], \
                   self.num_rotatable_bonds[item], self.polar_surface_area[item]
        else:
            return self.smiles[item], self.graphs[item], self.labels[item]



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Variant of Graph Convolutional Networks/Graph Isomorphism Networks
# for graph property prediction in OGB's Examples

import torch.nn as nn
import torch.nn.functional as F

from dgl.nn.pytorch.glob import SumPooling, AvgPooling, MaxPooling

from ..gnn.gnn_ogb import GNNOGB

__all__ = ['GNNOGBPredictor']

# pylint: disable=W0221, C0103
class GNNOGBPredictor(nn.Module):
    r"""Variant of GCN/GIN from `Open Graph Benchmark: Datasets for Machine Learning on Graphs
    <https://arxiv.org/abs/2005.00687>`__ for graph property prediction

    Parameters
    ----------
    in_edge_feats : int
        Number of input edge features.
    num_node_types : int
        Number of node types to embed. (Default: 1)
    hidden_feats : int
        Size for hidden representations. (Default: 300)
    n_layers : int
        Number of GNN layers to use. (Default: 5)
    n_tasks : int
        Number of output tasks. (Default: 1)
    batchnorm : bool
        Whether to apply batch normalization. (Default: True)
    activation : callable or None
        Activation function to apply to the output of each GNN layer except for the last layer.
        If None, no activation will be applied. (Default: ReLU)
    dropout : float
        The probability for dropout. (Default: 0, i.e. no dropout)
    gnn_type : str
        The GNN type to use, which can be either 'gcn' or 'gin'. (Default: 'gcn')
    virtual_node : bool
        Whether to use virtual node. (Default: True)
    residual : bool
        Whether to apply residual connections for virtual node embeddings. (Default: False)
    jk : bool
        Whether to sum over the output of all GNN layers as in
        `JK networks <https://arxiv.org/abs/1806.03536>`__. (Default: False)
    readout : str
        The readout function for computing graph-level representations out of node
        representations, which can be 'mean', 'sum' or 'max'. (Default: 'mean')
    """
    def __init__(self,
                 in_edge_feats,
                 num_node_types=1,
                 hidden_feats=300,
                 n_layers=5,
                 n_tasks=1,
                 batchnorm=True,
                 activation=F.relu,
                 dropout=0.,
                 gnn_type='gcn',
                 virtual_node=True,
                 residual=False,
                 jk=False,
                 readout='mean'):
        super(GNNOGBPredictor, self).__init__()

        assert gnn_type in ['gcn', 'gin'], \
            "Expect gnn_type to be 'gcn' or 'gin', got {}".format(gnn_type)
        assert readout in ['mean', 'sum', 'max'], \
            "Expect readout to be in ['mean', 'sum', 'max'], got {}".format(readout)

        self.gnn = GNNOGB(in_edge_feats=in_edge_feats,
                          num_node_types=num_node_types,
                          hidden_feats=hidden_feats,
                          n_layers=n_layers,
                          batchnorm=batchnorm,
                          activation=activation,
                          dropout=dropout,
                          gnn_type=gnn_type,
                          virtual_node=virtual_node,
                          residual=residual,
                          jk=jk)

        if readout == 'mean':
            self.readout = AvgPooling()
        if readout == 'sum':
            self.readout = SumPooling()
        if readout == 'max':
            self.readout = MaxPooling()

        self.predict = nn.Linear(hidden_feats, n_tasks)

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.gnn.reset_parameters()
        self.predict.reset_parameters()

    def forward(self, g, node_feats, edge_feats):
        """Predict graph properties.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : LongTensor of shape (N, 1)
            Input categorical node features. N for the number of nodes.
        edge_feats : FloatTensor of shape (E, in_edge_feats)
            Input edge features. E for the number of edges.

        Returns
        -------
        FloatTensor of shape (B, n_tasks)
            Predicted graph properties
        """
        node_feats = self.gnn(g, node_feats, edge_feats)
        graph_feats = self.readout(g, node_feats)

        return self.predict(graph_feats)
